from canvas_sak.core import *

@canvas_sak.command()
@click.argument("course")
@click.argument("csv_output_file", type=click.File("w"))
def export_letter_grade(course, csv_output_file):
    ''' export course letter grade to CSV

    the "Reported Letter Grade" column must be setup in the gradebook.
    this command will pull down the letter grades from that column an print a CSV record
    with the student id and the corresponding letter grade.
    output will got to the indicated csv_output_file.
    an output file name of - will go to stdout.
    '''

    canvas = get_canvas_object()
    course = get_course(canvas, course)

    rlg_assignment = get_assignment(course, "Reported Letter Grade")
    if not rlg_assignment:
        error('the "Reported Letter Grade" assignment hasn\'t been set up')
        exit(2)

    count = 0
    csv_output_file.write("Student ID,Grade\n")
    for submission in rlg_assignment.get_submissions():
        user = course.get_user(submission.user_id)
        if user.sis_user_id:
            csv_output_file.write(f"{user.sis_user_id}, {submission.grade}\n")
            count += 1

    info(f"{count} records written to {csv_output_file.name}")




from canvas_sak.core import *
from canvas_sak.core import filter_ignored_paths
import glob

@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('assignment_name', metavar='assignment', default='')
@click.argument('language', metavar='language')
@click.option('--dryrun/--no-dryrun', default=True, show_default=True, help="only show the grade, don't actually set it")
@click.option('--pause/--no-pause', default=False, show_default=True, help="pause before uploading")
@click.option('--multiple/--no-multiple', default=False, show_default=True, help="collect submissions from multiple classes")
def code_similarity(course_name, language, assignment_name, dryrun, pause, multiple):
    '''
    check submissions for code similarity using stanford MOSS.
    '''
    canvas = get_canvas_object()
    courses = get_courses(canvas, course_name)
    if len(courses) == 0:
        error(f"no courses matched {course_name}")
        exit(2)
    if len(courses) > 1 and not multiple:
        error(f"multiple matches for {course_name}")
        for course in courses:
            error(f"    {course.name}")
        exit(2)

    parser = ConfigParser()
    parser.read([config_ini])
    moss_userid = parser['MOSS']['userid']

    moss = mosspy.Moss(moss_userid, language)
    with tempfile.TemporaryDirectory("canvas_sak.attach") as tempdir:
        for course in courses:
            assignment = get_assignment(course, assignment_name)
            usermap = {u.id: u.name for u in course.get_users()}
            submissions = list(assignment.get_submissions())
            with click.progressbar(length=len(submissions), label="downloading", item_show_func=lambda x: x) as bar:
                for sub in assignment.get_submissions():
                    if sub.user_id not in usermap:
                        continue
                    udir = f"{tempdir}/{usermap[sub.user_id]}"
                    os.makedirs(udir)
                    bar.update(1, usermap[sub.user_id])
                    for attachment in sub.attachments:
                        aname = f"{udir}/{attachment.filename}"
                        with open(aname, "wb") as f:
                            f.write(requests.get(attachment.url).content)
                        if aname.endswith(".zip"):
                            with zipfile.ZipFile(aname, "r") as zf:
                                zf.extractall(udir)
        all_files = glob.glob(f"{tempdir}/**/*.{language}", recursive=True)
        # Filter out __MACOSX directories (common in zip files) and apply ignore patterns
        files_to_upload = [x for x in all_files if '/__MACOSX/' not in x]
        files_to_upload = filter_ignored_paths(files_to_upload)
        info(f"uploading {files_to_upload}")

        if pause:
            input(f"pausing. code is in {tempdir}. hit enter to continue")
        if dryrun:
            info(f"would upload {len(files_to_upload)} files to MOSS")
        else:
            for file in files_to_upload:
                moss.addFile(file)
            count = len(moss.files)
            with click.progressbar(length=count, label="uploading", item_show_func=lambda x: x) as bar:
                moss_url = moss.send(on_send=lambda fp, dn: bar.update(1, dn))
            info(f"results at {moss_url}")
            info(f"download with: wget -k -e robots=off -np -r {moss_url}")



from canvas_sak.core import *

@canvas_sak.command()
@click.argument("course")
@click.option('-m', 'min_grade', default=50.0, show_default=True, help="""
              the minimum assignment grade. any score below this grade will be set to
              this minimum score.
              """)
def min_grade_analyzer(course, min_grade):
    '''see what the scores would look like with minimum grade'''
    canvas = get_canvas_object()
    min_grade = min_grade / 100
    for course in get_courses(canvas, course, is_active=False, is_finished=True):
        # first get all the grade categories and track the ones with weights
        results = canvas.graphql('query { course(id: "' + str(course.id) + '''") {
                    enrollmentsConnection {
                        nodes {
                            grades { currentScore }
                            user { name }
                        }
                    }
                    assignmentGroupsConnection {
                      nodes {
                        groupWeight
                        name
                        id
                      }
                    }
                  }
                } ''')
        class_grade_by_student = {}
        for enrollment in results['data']['course']['enrollmentsConnection']['nodes']:
            class_grade_by_student[enrollment['user']['name']] = enrollment['grades']['currentScore']

        grades_by_student = defaultdict(lambda: defaultdict(list))
        assignment_groups = [assignment_group for assignment_group in
                             results['data']['course']['assignmentGroupsConnection']['nodes'] if
                             assignment_group['groupWeight']]
        weights = {}
        for assignment_group in assignment_groups:
            category = assignment_group['name']
            weight = assignment_group['groupWeight']
            weights[category] = weight
            assignment_group_id = assignment_group['id']
            assignments = canvas.graphql('query { assignmentGroup(id: "' + assignment_group_id + '''") {
                                         assignmentsConnection {
                                             nodes {
                                                 id
                                                 name
                                             }
                                         }}}''')
            for assignment in assignments['data']['assignmentGroup']['assignmentsConnection']['nodes']:
                assignment_id = assignment['id']
                scores = canvas.graphql('query { assignment(id: "' + assignment_id + '''") {
                                             name
                                             pointsPossible
                                             submissionsConnection {
                                                 nodes {
                                                     score
                                                      user { name }
                                                 }
                                             }
                                         } }''')
                points_possible = scores['data']['assignment']['pointsPossible']
                for score in scores['data']['assignment']['submissionsConnection']['nodes']:
                    currentScore = score['score']
                    name = score['user']['name']
                    if currentScore == None:
                        continue
                    grades_by_student[name][category].append((currentScore, points_possible))

        for (name, assignments) in grades_by_student.items():
            total = 0.0
            min_total = 0.0
            components = []
            for (cat, scores) in assignments.items():
                cat_total = sum([current_score for (current_score, points_possible) in scores])
                min_scores = [(
                    current_score if not points_possible or current_score >= points_possible * min_grade else points_possible * min_grade,
                    points_possible) for (current_score, points_possible) in scores]
                min_cat_total = sum([current_score for (current_score, points_possible) in min_scores])
                cat_possible = sum([points_possible for (current_score, points_possible) in scores])
                if cat_possible == 0:
                    cat_possible = 100
                cat_avg = cat_total / cat_possible
                min_avg = min_cat_total / cat_possible
                inc = cat_avg * weights[cat]
                min_inc = min_avg * weights[cat]
                total += inc
                min_total += min_inc
                components.append((cat,cat_avg * 100))
                # print(f'{scores} {cat_avg} {weights[cat]} {inc} {total} {min_total} {" " if total == min_total else "*****"}')
            letter = to_letter_grade(total)
            min_letter = to_letter_grade(min_total)
            if letter != min_letter:
                output(f'{name}@{class_grade_by_student[name]}@{total}({letter}) {min_total}({min_letter})')




from canvas_sak.core import *

@canvas_sak.command()
@click.argument("course")
@click.option("--round", default=0.0, help="points to add to the final score before calculating the letter grade.")
@click.option("--dryrun/--no-dryrun", default=True)
@click.option("--skip-mismatch/--no-skip-mismatch", default=True, help="do not set letter grade for current grades that don't match total")
def set_letter_grade(course, round, dryrun, skip_mismatch):
    ''' calculate the letter grade based on the final score in the class.

    the "Reported Letter Grade" assignment must be created in the gradebook as a letter grade assignment
    before this command is run.
    the command will loop through all the students in the class and set the letter grade in that assignment
    based on the final score in the class.
    '''

    canvas = get_canvas_object()
    course = get_course(canvas, course)

    rlg_assignment = get_assignment(course, "Reported Letter Grade")
    if not rlg_assignment:
        error('the "Reported Letter Grade" assignment hasn\'t been set up')
        exit(2)

    user_to_grade = {}
    for enrollment in course.get_enrollments(include=['grades']):
        if hasattr(enrollment, "grades"):
            current_score = enrollment.grades['current_score']
            final_score = enrollment.grades['final_score']
            if current_score != final_score:
                mess = f"current_score of {current_score} != {final_score} for {enrollment.user['name']} "
                if skip_mismatch:
                    warn(mess + "SKIPPED")
                    continue
                else:
                    warn(mess + "NOT SKIPPED")
            letter = points_to_letter(enrollment.grades['final_score'], round)
            user_to_grade[enrollment.user['id']] = (enrollment.user, letter, enrollment.grades['final_score'])

    if dryrun:
        for submission in rlg_assignment.get_submissions():
            if submission.user_id in user_to_grade:
                (user, letter, score) = user_to_grade[submission.user_id]
                info(f"{letter} {score} {user['name']}")
        warn("This was a dryrun. Nothing has been updated")
    else:
        with click.progressbar(length=len(user_to_grade), label="updating grades", show_pos=True) as bar:
            for submission in rlg_assignment.get_submissions():
                if submission.user_id in user_to_grade:
                    (user, letter, score) = user_to_grade[submission.user_id]
                    submission.edit(submission={'posted_grade': letter})
                    bar.update(1)



from canvas_sak.core import *

@canvas_sak.command()
def help_me_setup():
    """provide guidance through the setup process"""
    if os.path.isfile(config_ini):
        info(f"great! {config_ini} exists. let's check it!")
    else:
        error(f"""{config_ini} does not exist. you need to create it.
it should have the form:""")
        print_config_ini_format(True)
        sys.exit(2)

    parser = ConfigParser()
    try:
        parser.read([config_ini])
    except:
        error(f"there was a problem reading {config_ini}. make sure it has the format of:")
        print_config_ini_format(False)

    check_key("SERVER", parser)
    url = check_key("url", parser["SERVER"])
    p = urllib.parse.urlparse(url)
    if p.scheme != "https":
        error(f"url in {config_ini} must start with https://")
        sys.exit(2)

    if p.path:
        error(f"url in {config_ini} must have the form http://hostname with no other /")
        sys.exit(2)

    try:
        with urllib.request.urlopen(url) as con:
            info(f"{url} is reachable.")
    except Exception as e:
        error(f"got '{e}' accessing {url}. please check the url in {config_ini}.")
        sys.exit(2)

    token = check_key("token", parser["SERVER"])
    if token and len(token) > 20:
        info(f"token found. checking to see if it is usable")
    else:
        error(f"token is too short. make sure you have copied it correctly from canvas.")
        sys.exit(2)

    try:
        canvas = Canvas(url, token)
        info(f"you are successfully able to use canvas as {canvas.get_current_user().name}")
    except Exception as e:
        error(f"problem accessing canvas: {e}")
        sys.exit(2)

    sys.exit(0)



from canvas_sak.core import *

@canvas_sak.command()
@click.argument('course')
@click.argument('subject')
@click.option('--course-in-subject/--no-course-in-subject', show_default=True, default=True,
              help='include the course name in []s in the subject line')
@click.option('--message', help="message to send")
@click.option('--from-file', help="file containing message to send (- for stdin)", type=click.File('r'))
@click.argument('students', nargs=-1, required=True)
def message_students(course, subject, message, course_in_subject, from_file, students):
    '''message students in a course'''
    canvas = get_canvas_object()
    course = get_course(canvas, course)

    message_to_send = ""
    if message:
        if not message.endswith("\n"):
            message += "\n"
        message_to_send += message
    if from_file:
        message_to_send += from_file.read()
    if not message and not from_file:
        error("either --message or --from_file must be specified")
        sys.exit(2)

    if course_in_subject:
        subject = f'[{format_course_name(course.name)}] {subject}'

    found_error = False
    to_message = []
    for student in students:
        users = list(course.get_users(search_term=student))
        if not len(users):
            error(f"could not find {student}")
            found_error = True
            continue
        elif len(users) > 1:
            error(f"multiple matches for {student}: {', '.join([str(u) for u in users])}")
            found_error = True
            continue
        to_message.append(users[0])
    if found_error:
        sys.exit(2)

    for user in to_message:
        # we set group_conversation to true to make sure it shows up as a new conversation
        canvas.create_conversation([user.id], message_to_send, subject=subject, group_conversation=True)



import time

import click

from canvas_sak.core import *


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('qti_file', type=click.Path(exists=True))
@click.option('--wait/--no-wait', default=True, show_default=True,
              help="wait for the import to complete")
@click.option('--poll-interval', default=2, show_default=True,
              help="seconds between status checks when waiting")
def upload_qti_quiz(course_name, qti_file, wait, poll_interval):
    """Upload a QTI quiz package to a Canvas course.

    QTI_FILE should be a zip file containing a QTI-formatted quiz.
    """
    canvas = get_canvas_object()
    course = get_course(canvas, course_name)
    output(f"uploading QTI quiz to {course.name}")

    # Create content migration for QTI import
    migration = course.create_content_migration(
        migration_type='qti_converter',
        pre_attachment={
            'name': os.path.basename(qti_file),
        }
    )

    # Upload the QTI file
    info(f"uploading {qti_file}...")
    upload_url = migration.pre_attachment['upload_url']
    upload_params = migration.pre_attachment['upload_params']

    with open(qti_file, 'rb') as f:
        files = {'file': (os.path.basename(qti_file), f)}
        response = requests.post(upload_url, data=upload_params, files=files)

    if response.status_code not in [200, 201, 301, 303]:
        error(f"failed to upload file: {response.status_code} {response.text}")
        sys.exit(1)

    info("file uploaded, processing migration...")

    if wait:
        # Poll for migration completion
        while True:
            migration = course.get_content_migration(migration.id)
            status = migration.workflow_state

            if status == 'completed':
                info("quiz import completed successfully")
                break
            elif status == 'failed':
                error(f"quiz import failed: {getattr(migration, 'migration_issues', 'unknown error')}")
                sys.exit(1)
            elif status in ['pre_processing', 'pre_processed', 'running', 'queued', 'exporting']:
                output(f"status: {status}...")
                time.sleep(poll_interval)
            else:
                warn(f"unexpected status: {status}")
                time.sleep(poll_interval)
    else:
        output(f"migration started with id {migration.id}")
        output("use Canvas to check migration status")



import csv
from importlib.metadata import requires

from canvas_sak.core import *

@canvas_sak.command()
@click.argument("course")
@click.argument("assignment")
@click.option("--file", type=click.File('r'), required=True)
@click.option("--id", required=True, help="field number (0-based) in the file that contains the ID")
@click.option("--grade", required=True, help="field number (0-based) in the file that contains the grade")
@click.option("--free-points", default=0.0, help="points to add to the score.")
@click.option("--dryrun/--no-dryrun", default=True)
def upload_assignment_grades(course, assignment, file, id, grade, free_points, dryrun):
    ''' upload grades for an assignment from a CSV file.
    '''

    canvas = get_canvas_object()
    course = get_course(canvas, course)

    assignment = get_assignment(course, assignment)
    if not assignment:
        error(f'{assignment} hasn\'t been set up')
        exit(2)

    grades = {}
    csv_reader = csv.reader(file)
    next(csv_reader)  # skip header
    for row in csv_reader:
        student_id = row[int(id)].strip()
        student_grade = row[int(grade)].strip()
        if not student_id or not student_grade:
            print(f"skipping empty id or grade: '{student_id}' '{student_grade}' in {row}")
        student_grade = student_grade.rstrip("%")
        student_grade = min(100.00, float(student_grade) + free_points)
        grades[student_id] = student_grade

    user_id_2_sis = {}
    for enrollment in course.get_enrollments(include = ["user", "sis_user_id"]):
        user_id_2_sis[enrollment.user['id']] = enrollment.sis_user_id

    if dryrun:
        for submission in assignment.get_submissions():
            sis = user_id_2_sis[submission.user_id] if submission.user_id in user_id_2_sis else None
            if sis in grades:
                score = grades[sis]
                info(f"{score} {sis}")
            else:
                warn(f"no grade found for {sis}")
        warn("This was a dryrun. Nothing has been updated")
    else:
        with click.progressbar(length=len(grades), label="updating grades", show_pos=True) as bar:
            for submission in assignment.get_submissions():
                sis = user_id_2_sis[submission.user_id] if submission.user_id in user_id_2_sis else None
                if sis in grades:
                    score = grades[sis]
                    submission.edit(submission={'posted_grade': score})
                    bar.update(1)
                else:
                    warn(f"no grade found for {sis}")



import datetime
import fnmatch
import functools
import logging
import os
import re
import string
import sys
import tempfile
import urllib
import urllib.parse
import urllib.request
import zipfile
from collections import defaultdict, namedtuple
from configparser import ConfigParser
from html.parser import HTMLParser
from typing import NamedTuple

import canvasapi.file
import click
import mosspy
import requests
from canvasapi import Canvas
from canvasapi.course import Course
from canvasapi.discussion_topic import DiscussionEntry
from canvasapi.requester import Requester

course_name_matcher = r"((\S*): (\S+)\s.*)"
course_name_formatter = r"\2:\3"


def format_course_name(name, matcher=course_name_matcher, formatter=course_name_formatter):
    if formatter == '-':
        return name
    return re.sub(matcher, formatter, name)


def introspect(o):
    print("class", o.__class__)
    for i in dir(o):
        print(i)


# this file has the url and token we will use
config_ini = click.get_app_dir("canvas_sak.ini")

# local ignore patterns file name
IGNORE_FILE = "canvas-sak-ignore.lst"


def load_ignore_patterns():
    """Load ignore patterns from config file [IGNORE] section and local canvas-sak-ignore.lst file.

    Returns a list of gitignore-style patterns.
    """
    patterns = []

    # Load from config file [IGNORE] section
    parser = ConfigParser()
    parser.read([config_ini])
    if "IGNORE" in parser:
        for key, value in parser["IGNORE"].items():
            # Each value in the section is a pattern
            if value:
                patterns.append(value)
            # The key itself can also be a pattern (for key-only entries)
            elif key:
                patterns.append(key)

    # Load from local canvas-sak-ignore.lst file
    if os.path.exists(IGNORE_FILE):
        with open(IGNORE_FILE, "r") as f:
            for line in f:
                line = line.strip()
                # Skip empty lines and comments
                if line and not line.startswith("#"):
                    patterns.append(line)

    return patterns


def matches_ignore_pattern(path, patterns):
    """Check if a path matches any of the ignore patterns.

    Supports gitignore-style patterns:
    - * matches any sequence of characters (except /)
    - ** matches any sequence of characters including /
    - ? matches any single character
    - patterns ending with / only match directories
    - patterns starting with ! negate the match

    Args:
        path: The file path to check (should use forward slashes)
        patterns: List of gitignore-style patterns

    Returns:
        True if the path should be ignored, False otherwise.
    """
    # Normalize path to use forward slashes
    path = path.replace("\\", "/")

    # Track if path is ignored (can be toggled by negation patterns)
    ignored = False

    for pattern in patterns:
        if not pattern:
            continue

        # Handle negation patterns
        negation = pattern.startswith("!")
        if negation:
            pattern = pattern[1:]

        # Handle directory-only patterns (ending with /)
        dir_only = pattern.endswith("/")
        if dir_only:
            pattern = pattern[:-1]
            # For dir-only patterns, we need to check if path is a directory
            # Since we're dealing with file paths, we check if any path component matches
            # For simplicity, we'll match if the pattern appears as a path segment

        # Handle ** patterns (recursive matching)
        if "**" in pattern:
            # Convert pattern to regex using placeholder to preserve **
            DOUBLESTAR = "\x00DOUBLESTAR\x00"
            regex_pattern = pattern.replace("**/", DOUBLESTAR)  # **/ matches zero or more dirs
            regex_pattern = regex_pattern.replace("**", DOUBLESTAR)  # remaining **
            # Escape special regex chars
            regex_pattern = re.escape(regex_pattern)
            # Restore ** as pattern that matches zero or more path segments
            regex_pattern = regex_pattern.replace(DOUBLESTAR, "(.*/)?")
            # Convert remaining * and ? to regex equivalents
            regex_pattern = regex_pattern.replace(r"\*", "[^/]*")
            regex_pattern = regex_pattern.replace(r"\?", "[^/]")
            regex_pattern = f"^{regex_pattern}($|/)" if dir_only else f"^{regex_pattern}$"

            if re.search(regex_pattern, path):
                ignored = not negation
        else:
            # Handle patterns without **
            # Check if pattern should match from root or anywhere
            if "/" in pattern:
                # Pattern with / matches from the beginning or specific path
                if pattern.startswith("/"):
                    # Anchored to root
                    pattern = pattern[1:]
                    if fnmatch.fnmatch(path, pattern):
                        ignored = not negation
                else:
                    # Match anywhere in path
                    if fnmatch.fnmatch(path, pattern) or fnmatch.fnmatch(path, f"*/{pattern}"):
                        ignored = not negation
            else:
                # Pattern without / matches filename or any path component
                filename = os.path.basename(path)
                if fnmatch.fnmatch(filename, pattern):
                    ignored = not negation
                # Also check if pattern matches any path component
                for component in path.split("/"):
                    if fnmatch.fnmatch(component, pattern):
                        ignored = not negation
                        break

    return ignored


def filter_ignored_paths(paths, patterns=None):
    """Filter a list of paths, removing those that match ignore patterns.

    Args:
        paths: Iterable of file paths to filter
        patterns: Optional list of patterns. If None, loads patterns using load_ignore_patterns().

    Returns:
        List of paths that don't match any ignore patterns.
    """
    if patterns is None:
        patterns = load_ignore_patterns()

    if not patterns:
        return list(paths)

    return [p for p in paths if not matches_ignore_pattern(p, patterns)]


def error(message):
    click.echo(click.style(message, fg='red'))


def info(message):
    click.echo(click.style(message, fg='blue'))


def warn(message):
    click.echo(click.style(message, fg='yellow'))


def output(message):
    click.echo(message)


@functools.lru_cache
def get_requester():
    parser = ConfigParser()
    parser.read([config_ini])
    if "SERVER" not in parser:
        error(f"did not find [SERVER] section in {config_ini}")
        info("try using the help-me-setup command")
        sys.exit(1)
    return Requester(parser['SERVER']['url'], parser['SERVER']['token'])


access_token = None
canvas_url = None


def get_canvas_object():
    parser = ConfigParser()
    parser.read([config_ini])
    if "SERVER" not in parser:
        error(f"did not find [SERVER] section in {config_ini}")
        info("try using the help-me-setup command")
        sys.exit(1)
    if 'url' not in parser['SERVER'] or 'token' not in parser['SERVER']:
        error(f"did not find url or token in {config_ini}")
        info("try using the help-me-setup command")
        sys.exit(1)
    try:
        canvas = Canvas(parser['SERVER']['url'], parser['SERVER']['token'])
        user = canvas.get_current_user()
        info(f"accessing canvas as {user.name} ({user.id})")
        canvas.user_id = user.id
        global access_token, canvas_url
        access_token = parser['SERVER']['token']
        canvas_url = parser['SERVER']['url']
        return canvas
    except:
        error(f"there was a problem accessing canvas. try using help-me-setup.")
        sys.exit(2)


@click.group()
@click.option("--log-level", type=click.Choice(['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'], case_sensitive=False),
              help="set python logging level")
def canvas_sak(log_level):
    if log_level:
        log_level_int = getattr(logging, log_level.upper())
        logging.basicConfig(level=log_level_int)


def get_course(canvas, name, is_active=True) -> Course:
    """ find one course based on partial match """
    course_list = get_courses(canvas, name, is_active)
    if len(course_list) == 0:
        error(f'no courses found that contain {name}. options are:')
        for c in get_courses(canvas, "", is_active):
            error(fr"    {c.name}")
        sys.exit(2)
    elif len(course_list) > 1:
        error(f"multiple courses found for {name}:")
        for c in course_list:
            error(f"    {c.name}")
        sys.exit(2)
    return course_list[0]


def get_courses(canvas: Canvas, name: str, is_active=True, is_finished=False) -> [Course]:
    ''' find the courses based on partial match '''
    courses = canvas.get_courses(enrollment_type="teacher")
    now = datetime.datetime.now(datetime.timezone.utc)
    course_list = []
    for c in courses:
        start = c.start_at_date if hasattr(c, "start_at_date") else now
        end = c.end_at_date if hasattr(c, "end_at_date") else now
        if is_active and (start > now or end < now):
            continue
        if is_finished and end >= now:
            continue
        if name in c.name:
            c.start = start
            c.end = end
            course_list.append(c)
    return course_list


def get_assignments(course, title):
    ''' find the assignment based on partial match '''
    assignments = list(course.get_course_level_assignment_data())
    filtered_assignments = [a for a in assignments if title in a['title']]
    if not title:
        # if an assignment title wasn't specified, we don't want to return
        # anything even if there is one result.
        output(f'possible assignments for {course.name}')
        for a in assignments:
            output(f"    {a['title']}")
        sys.exit(1)
    return filtered_assignments


def get_assignment(course, title):
    filtered_assignments = get_assignments(course, title)
    if len(filtered_assignments) == 0:
        error(f'{title} assignment not found. possible assignments are:')
        for a in get_assignments(course, ""):
            error(f"    {a['title']}")
        sys.exit(2)
    if len(filtered_assignments) > 1:
        # sometimes there are prefix matches for an assignment name that is
        # fully given: "Assignment 1" and "Assignment 1 Extended"
        # if there are multiple matches but one exact match, use the exact match
        exact_match = [a for a in filtered_assignments if title == a['title']]
        if len(exact_match) == 1:
            filtered_assignments = exact_match
        else:
            error(f'multiple assignments found matching {title}:')
            for a in filtered_assignments:
                error(f"    {a['title']}")
            sys.exit(2)
    return course.get_assignment(filtered_assignments[0]['assignment_id'])


def maybe_a_word(word):
    if not word.isalpha():
        return False

    word = word.strip().lower()
    vowels = [x for x in word if x in "aeiou"]
    if not vowels:
        return False

    ratio = round(len(word) / len(vowels), 1)
    return 1.5 <= ratio <= 8.0


class WordCounter(HTMLParser):
    def __init__(self):
        super().__init__()
        self.checked_word_count = 0
        self.word_count = 0

    def handle_data(self, data):
        data.translate(str.maketrans('', '', string.punctuation))
        words = re.findall(r'\w+', data)
        self.word_count += len(words)
        self.checked_word_count += len(set([w for w in words if maybe_a_word(w)]))


def count_words(content):
    wc = WordCounter()
    wc.feed(content)
    return wc.checked_word_count

def print_config_ini_format(is_info):
    func = info if is_info else error
    func("""[SERVER]
url=https://XXXXX.instructure.com
token=YYYYYYYYYYYYYYYYYYYYYYY
[MOSS]
userid=ZZZZZZZZZXX

where XXXXX is your organization identifier. for example, SJSU is sjsu.
YYYYYYYYYYYYYYYYYYYYY is the token you generate in canvas with Account->Settings->[New Access Token]
it will look like a long list of letters, numbers, and symbols. copy them after token=
""")


def check_key(key, obj):
    if key not in obj:
        error(f"{key} not in {config_ini}. make sure it has the format of:")
        print_config_ini_format(False)
        sys.exit(2)
    return obj[key]


# very simple sanitizer to escape semicolons, tabs, and newlines
def sanitize(name):
    return name.replace(';', ',').replace("\n", "\\n").replace("\t", "\\t")


ResourceRecord = namedtuple("ResourceRecord", ["id", "url", "type", "name", "stub"])

# rr4name and rr4id will have the ResourceRecord type prepended
rr4name = {}
rr4id = {}
rr4url = {}
course_modules = {}


def process_resource_record(rr):
    rr4name[rr.type + rr.name] = rr
    rr4id[rr.type + str(rr.id)] = rr
    rr4url[rr.url] = rr


# strip any query params off (in different places the same url will have different params
def base_url(url):
    return url.split("?")[0]


def map_course_resource_records(course):
    with click.progressbar(length=6, label="mapping existing resources") as bar:
        for folder in course.get_folders():
            for file in folder.get_files():
                process_resource_record(
                    ResourceRecord(file.id, base_url(file.url), "File", os.path.join(str(folder), str(file)).replace("\\", "/"), file.size == 0))
        bar.update(1)
        for assignment in course.get_assignments():
            process_resource_record(
                ResourceRecord(assignment.id, base_url(assignment.html_url), "Assignment", assignment.name, not assignment.description))
        bar.update(1)
        for discussion in course.get_discussion_topics():
            process_resource_record(
                ResourceRecord(discussion.id, base_url(discussion.html_url), "Discussion", discussion.title, not discussion.message))
        bar.update(1)
        for page in course.get_pages(include=["body"]):
            process_resource_record(ResourceRecord(page.page_id, base_url(page.url), "Page", page.title, not page.body))
        bar.update(1)
        for quiz in course.get_quizzes():
            process_resource_record(ResourceRecord(quiz.id, base_url(quiz.html_url), "Quiz", quiz.title, not quiz.description))
        bar.update(1)
        for mod in course.get_modules():
            course_modules[mod.name] = mod
        bar.update(1)

letter_grades = [(96, "A+"), (93, "A"), (90, "A-"), (86, "B+"), (83, "B"), (80, "B-"), (76, "C+"), (73, "C"),
                 (70, "C-"), (66, "D+"), (63, "D"), (60, "D-"), (0, "F")]

def to_letter_grade(score):
    if score > 89:
        return 'A'
    if score > 79:
        return 'B'
    if score > 69:
        return 'C'
    if score > 59:
        return 'D'
    return 'F'

def points_to_letter(points, round):
    if not points:
        return 'WU'
    points += round
    for letter in letter_grades:
        if points >= letter[0]:
            return letter[1]



from canvas_sak.core import *

@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('assignment_name', metavar='assignment', default='')
@click.option('--dryrun/--no-dryrun', default=True, show_default=True,
              help="only show the grade, don't actually set it")
@click.option('--min-words', default=5, show_default=True, help="the minimum number of valid words to get credit")
@click.option('--points-comment', default=1, show_default=True, help="number of points for posting a comment")
@click.option('--max-points', default=2, show_default=True, help="maximum number of points to give")
def grade_discussion(course_name, assignment_name, dryrun, min_words, points_comment, max_points):
    '''
    grade a discussion assignment based on participation.

    one point is added for a post and another for a reply for a total of 2.
    this tool assumes that the student must post first to reply.

    course_name - any part of an active course name. for example, 249 will match CS249.
    the course must active (it has not passed the end date) to be eligible for matching.
    only the first match will be used.

    assignment_name - any part of an assigment's name will be matched. only the first
    match will be used.
    '''

    canvas = get_canvas_object()

    course = get_course(canvas, course_name)

    now = datetime.datetime.now(datetime.timezone.utc)
    assignments = get_assignments(course, assignment_name)
    for assignment_data in assignments:
        info(f"grading {assignment_data['title']}")
        assignment = course.get_assignment(assignment_data['assignment_id'])
        due_at_date = assignment.due_at_date
        if due_at_date > now:
            warn(f"{assignment_data['title']} not due: skipping")
            continue
        submissions = assignment.get_submissions()
        grades = {}
        skipped = 0
        processed = 0

        # we calculate the grades and upload them in two steps since the second
        # step is slow so we want to do a progress bar. we could still do it all
        # in one pass if we could get the count. i can't find a way to do that
        # without going through the submissions
        for s in submissions:
            processed += 1
            if not hasattr(s, "discussion_entries"):
                skipped += 1
                continue

            grade = 0
            for entry in s.discussion_entries:
                entry = DiscussionEntry(None, entry)
                if entry.created_at_date > due_at_date:
                    info(
                        f"skipping discussion from {s.user_id} submitted at {entry.created_at_date} but due {due_at_date}")
                    continue
                if count_words(entry.message) > 5:
                    grade = min(grade + points_comment, max_points)

            grades[s] = grade

        if skipped == processed:
            error(f"'{assignment.name}' doesn't appear to be a discussion assignment")
            sys.exit(2)

        info(f"processed {processed}, skipped {skipped}.")

        if dryrun:
            info("would have posted:")
            for i in grades.items():
                info(f"    {i[0]} {i[1]}")
        else:
            with click.progressbar(length=len(grades), label="updating grades", show_pos=True) as bar:
                for i in grades.items():
                    i[0].edit(submission={'posted_grade': i[1]})
                    bar.update(1)



from canvas_sak.commands.upload_canvas_course import page_name_to_url
from canvas_sak.core import *
from canvas_sak.md2fhtml import *


def download_modules(course, target, dryrun):
    def base_inner_module_to_str(module_item):
        return f'{"  " * (module_item.indent + 1)}* {sanitize(module_item.title)}; {module_item.type}{"" if module_item.published else "; published=False"}'

    def named_inner_module_to_str(module_item):
        module_item_target_name = None
        if hasattr(module_item, 'content_id'):
            module_item_target_name = rr4id[module_item.type + str(module_item.content_id)].name
        else:
            module_item_target_name = rr4url[module_item.page_url].name

        if module_item_target_name == module_item.title:
            module_item_target_name = None

        return f'{base_inner_module_to_str(module_item)}{f"; target={module_item_target_name}" if module_item_target_name else ""}'

    def external_tool(mi):
        return f'{base_inner_module_to_str(mi)}; newtab={"True" if mi.new_tab else "False"}; url={mi.external_url}'

    module_renderers = {
        "Assignment": named_inner_module_to_str,
        "File": named_inner_module_to_str,
        "Page": named_inner_module_to_str,
        "Quiz": named_inner_module_to_str,
        "Discussion": named_inner_module_to_str,
        "SubHeader": base_inner_module_to_str,
        "ExternalUrl": external_tool,
        "ExternalTool": external_tool,
    }

    top_modules = []
    id2name = {}
    output = ''
    for module in course.get_modules():
        id2name[module.id] = module.name
        ms = f"# {module.name}"
        if module.unlock_at:
            ms += f"; unlock={module.unlock_at}"
        if module.require_sequential_progress:
            ms += f"; sequential"
        if module.prerequisite_module_ids:
            ms += f"; prereqs={','.join([id2name[id] for id in module.prerequisite_module_ids])}"
        if hasattr(module, "completed_at") and module.completed_at:
            ms += f"; completed={module.completed_at}"
        if not module.published:
            ms += f"; published=False"
        output += ms + '\n'
        for item in module.get_module_items():
            if item.type in module_renderers:
                output += module_renderers[item.type](item) + '\n'
            else:
                warn(f"cannot render {item.__dict__}")

    if dryrun:
        info(f"would have written:\n{output}to {target}")
    else:
        with open(ensure_dirs(target), "w") as fd:
            fd.write(output)


def ensure_dirs(target):
    os.makedirs(os.path.dirname(target), exist_ok=True)
    return target

def download_discussions(course, target, dryrun):
    os.makedirs(target, exist_ok=True)
    for discussion in course.get_discussion_topics():
        # windows can't have : in the filename :'(
        target_file = os.path.join(target, discussion.title.strip().replace("\\", "-").replace(":", ";") + ".md")
        if os.path.exists(target_file):
            info(f"{target_file} already exists for {discussion.title}")
        else:
            if dryrun:
                info(f"would download {target_file} for {discussion.title}")
            else:
                info(f"downloading {target_file} for {discussion.title}")
                with open(ensure_dirs(target_file), "w+") as fd:
                    fd.write(f"# {discussion.title}\n")
                    # todo: we need to fix up the links based on the rr maps
                    fd.write(html2mdstr(discussion.message))


def download_assignments(course, target, dryrun):
    pass


def fix_links(text):
    # that funky ( [^\)]*) shouldn't be needed, but i do see trailing names after the URL
    return re.sub(r"\(https://\w+.instructure.com/courses/\w+/pages/([^ )]+)( [^\)]*)\)", r"(\1)", text)


def download_pages(course, target, dryrun):
    os.makedirs(target, exist_ok=True)
    for page in course.get_pages(include=["body"]):
        url = page_name_to_url(page.title)
        if page.url != url:
            warn(f"calculated page url for {page.title} ({url}) does not equal {page.url}")
        if dryrun:
            info(f"would download {page.title} to {url}")
        else:
            with open(ensure_dirs(os.path.join(target, url) + ".md"), "w+") as fd:
                fd.write(f"published: {page.published}\n")
                if page.publish_at:
                    fd.write(f"publish_at: {page.publish_at}\n")
                if page.front_page:
                    fd.write(f"front_page: {page.front_page}\n")
                fd.write(f"title: {page.title}\n")
                fd.write(fix_links(html2mdstr(page.body)))


def download_files(course, target, dryrun):
    class ToDownload(NamedTuple):
        file: canvasapi.file.File
        target: str

    error_seen = False
    to_download = []
    for folder in course.get_folders():
        target_dir = os.path.join(target, str(folder))
        if os.path.exists(target_dir):
            if not os.path.isdir(target_dir):
                error(f"{target_dir} is not a directory. skipping")
                error_seen = True
                continue
        else:
            if dryrun:
                info(f"would create {target_dir}")
            else:
                os.makedirs(target_dir)

        for file in folder.get_files():
            full_name = os.path.join(str(folder), str(file))
            target_file = os.path.join(target_dir, str(file))
            if dryrun:
                info(f"would download {full_name} to {target_file}")
            else:
                if os.path.exists(target_file):
                    warn(f"{target_file} already exists. skipping")
                else:
                    to_download.append(ToDownload(file, target_file))

    if to_download:
        with click.progressbar(to_download, label="downloading",
                               item_show_func=lambda i: str(i.file) if i else "") as tds:
            for td in tds:
                td.file.download(td.target)
    if error_seen:
        exit(2)


def download_announcements(course, target, dryrun):
    pass


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.option('--dryrun/--no-dryrun', default=True, show_default=True, help="show what would happen, but don't do it.")
@click.option('--modules/--no-modules', default=False, show_default=True,
              help=f"download modules to the {click.style('modules', underline=True, italic=True)} file.")
@click.option('--discussions/--no-discussions', default=False, show_default=True,
              help=f"download discussions to the {click.style('discussions', underline=True, italic=True)} subdirectory.")
@click.option('--assignments', default=False, show_default=True,
              help=f"download assignments to the {click.style('assignments', underline=True, italic=True)} subdirectory.")
@click.option('--pages/--no-pages', default=False, show_default=True,
              help=f"download pages to the {click.style('pages', underline=True, italic=True)} subdirectory.")
@click.option('--files', default=False, show_default=True,
              help=f"download files to the {click.style('files', underline=True, italic=True)} subdirectory.")
@click.option('--announcements', default=False, show_default=True,
              help=f"download announcements to the {click.style('announcements', underline=True, italic=True)} subdirectory.")
@click.option('--all/--no-all', default=False, show_default=True,
              help="download all content to corresponding directories")
@click.option("--target", default='.', show_default=True, help="download content parent directory.")
def download_course_content(course_name, dryrun, modules, discussions, assignments, pages, files, announcements, all,
                            target):
    """download course content from local files"""
    canvas = get_canvas_object()
    course = get_course(canvas, course_name, is_active=False)
    output(f"found {course.name}")
    map_course_resource_records(course)

    if all:
        modules = discussions = assignments = pages = files = announcements = True

    if not (modules or discussions or assignments or pages or files or announcements):
        error("nothing selected to download")
        exit(1)

    if modules:
        download_modules(course, os.path.join(target, 'modules'), dryrun)
    if discussions:
        download_discussions(course, os.path.join(target, 'discussions'), dryrun)
    if assignments:
        download_assignments(course, os.path.join(target, 'assignments'), dryrun)
    if pages:
        download_pages(course, os.path.join(target, 'pages'), dryrun)
    if files:
        download_files(course, os.path.join(target, 'files'), dryrun)
    if announcements:
        download_announcements(course, os.path.join(target, 'announcements'), dryrun)



from canvas_sak.core import *


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('image', metavar='image')
@click.option('--remove', is_flag=True, default=False, help="remove the course image instead of setting it")
def set_course_image(course_name, image, remove):
    """Set or remove the course image.

    IMAGE can be a local file path or a URL. If it's a local file, it will be
    uploaded to Canvas first. If it's a URL, it will be set directly.

    Examples:

        canvas-sak set-course-image "My Course" ./banner.jpg

        canvas-sak set-course-image "My Course" https://example.com/image.jpg

        canvas-sak set-course-image "My Course" --remove dummy
    """
    canvas = get_canvas_object()
    course = get_course(canvas, course_name, is_active=False)
    info(f"found {course.name}")

    if remove:
        info("removing course image")
        course.update(course={"remove_image": True})
        output("course image removed")
        return

    # Check if image is a URL or a file path
    if image.startswith("http://") or image.startswith("https://"):
        # It's a URL, set it directly
        info(f"setting course image to URL: {image}")
        course.update(course={"image_url": image})
        output("course image set from URL")
    else:
        # It's a file path, upload it first
        if not os.path.exists(image):
            error(f"file not found: {image}")
            sys.exit(1)

        info(f"uploading {image}")
        success, response = course.upload(image)
        if not success:
            error("failed to upload image")
            sys.exit(1)

        # Get the file ID from the response
        if isinstance(response, dict):
            file_id = response['id']
        else:
            file_id = response.id

        info(f"setting course image to uploaded file (id={file_id})")
        course.update(course={"image_id": file_id})
        output("course image set from uploaded file")



__all__ = [
    "archive_inbox",
    "code_similarity",
    "collect_reference_info",
    "download_canvas_course",
    "download_submissions",
    "export_letter_grade",
    "grade_discussion",
    "help_me_setup",
    "list_courses",
    "list_due_dates",
    "list_students",
    "message_students",
    "min_grade_analyzer",
    "set_course_image",
    "set_due_dates",
    "set_fudge_points",
    "set_letter_grade",
    "upload_assignment_grades",
    "upload_canvas_course",
    "upload_qti_quiz",
    "quiz",
]



from canvas_sak.core import *

@canvas_sak.command()
@click.option('--active/--inactive', default=True, help="show only active courses")
@click.option('--matcher', default=course_name_matcher, show_default=True, metavar="match_re_expression",
              help="course name regular expressions matcher")
@click.option('--formatter', default=course_name_formatter, show_default=True, metavar="format_re_expression", help="""
              course name regular expressions formatter based on the matcher pattern.
              a format pattern of - will turn off formatting.
              """)
def list_courses(active, matcher, formatter):
    '''list courses i am teaching. --inactive will include past and future courses.'''
    canvas = get_canvas_object()
    courses = get_courses(canvas, "", is_active=active)
    for c in courses:
        name = c.name if hasattr(c, "name") else "none"
        output(f"{c.id} {format_course_name(name, matcher, formatter)} {c.start:%Y-%m-%d} {c.end:%Y-%m-%d}")



from canvas_sak.core import *

@canvas_sak.command()
@click.argument('course_name')
@click.argument('quiz_name', default='')
@click.argument('points', default=-666, type=float)
@click.option('--dryrun/--no-dryrun', default=True, show_default=True,
              help="only show the grade, don't actually set it")
@click.option('--decrease/--no-decrease', default=False, show_default=True,
              help='If not true, the fudge points will not be updated if new points < old points.')
def set_fudge_points(course_name, quiz_name, points, dryrun, decrease):
    '''
    set the fudge points for a quiz.

    course_name - any part of an active course name. for example, 249 will match CS249.
    the course must active (it has not passed the end date) to be eligible for matching.
    only the first match will be used.

    quiz_name - any part of an quiz's name will be matched. if multiple quizes match, the
    points will not be set.
    '''

    canvas = get_canvas_object()

    course = get_course(canvas, course_name)

    quizzes = list(course.get_quizzes())
    selected_quizzes = [q for q in quizzes if quiz_name in q.title]

    if not selected_quizzes:
        error(f"could not find {quiz_name} in {', '.join([q.title for q in quizzes])}")
    elif len(selected_quizzes) > 1:
        error(f"multiple matches for {quiz_name}: {', '.join([q.title for q in selected_quizzes])}")
    else:
        for s in selected_quizzes[0].get_submissions():
            if points == -666:
                info(f"{s.user_id} {s.fudge_points}")
            else:
                if not decrease and s.fudge_points > points:
                    info(f"skipping {s.user_id} with {s.fudge_points} points")
                else:
                    if dryrun:
                        info(f"would update fudge points for {s.user_id} from {s.fudge_points} to {points}")
                    else:
                        info(f"updating fudge points for {s.user_id} from {s.fudge_points} to {points}")
                        # s.update_score_and_comments(fudge_points=points)
                        s.update_score_and_comments(quiz_submissions=[{"attempt": s.attempt, 'fudge_points': points}])



from canvas_sak.core import *


def format_date(dt_str):
    """Convert ISO date string to YYYY-MM-DD-hh:mm format in local timezone"""
    if not dt_str:
        return None
    dt = datetime.datetime.fromisoformat(dt_str.replace('Z', '+00:00'))
    local_dt = dt.astimezone()
    return local_dt.strftime('%Y-%m-%d-%H:%M')


def build_date_entries(unlock_at, due_at, lock_at):
    """Build comma-separated date entries string"""
    entries = []
    if unlock_at:
        formatted = format_date(unlock_at)
        if formatted:
            entries.append(f"available={formatted}")
    if due_at:
        formatted = format_date(due_at)
        if formatted:
            entries.append(f"due={formatted}")
    if lock_at:
        formatted = format_date(lock_at)
        if formatted:
            entries.append(f"until={formatted}")
    return ','.join(entries)


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.option('--active/--inactive', default=True, help="show only active courses")
def list_due_dates(course_name, active):
    """List due dates for all assignments in dates file format.

    Output format: assignment name TAB comma-separated dates

    Each date is type=YYYY-MM-DD-hh:mm where type is available, due, or until.

    Example: Homework 1\tavailable=2024-01-15-09:00,due=2024-01-22-23:59
    """
    canvas = get_canvas_object()
    course = get_course(canvas, course_name, active)

    for assignment in course.get_assignments():
        date_entries = build_date_entries(
            assignment.unlock_at,
            assignment.due_at,
            assignment.lock_at
        )
        output(f"{assignment.name}\t{date_entries}")



from canvas_sak.core import *


def parse_date(date_str):
    """Convert YYYY-MM-DD-hh:mm format (local time) to ISO format for Canvas API"""
    if not date_str:
        return None
    dt = datetime.datetime.strptime(date_str, '%Y-%m-%d-%H:%M')
    local_dt = dt.astimezone()
    return local_dt.isoformat()


def parse_date_entries(entries_str):
    """Parse comma-separated date entries into a dict"""
    result = {}
    if not entries_str.strip():
        return result
    for entry in entries_str.split(','):
        if '=' not in entry:
            continue
        key, value = entry.split('=', 1)
        key = key.strip()
        value = value.strip()
        if key == 'available':
            result['unlock_at'] = parse_date(value)
        elif key == 'due':
            result['due_at'] = parse_date(value)
        elif key == 'until':
            result['lock_at'] = parse_date(value)
    return result


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('dates_file', type=click.File('r'))
@click.option('--active/--inactive', default=True, help="show only active courses")
@click.option('--dryrun/--no-dryrun', default=True, show_default=True, help="show what would happen, but don't do it")
def set_due_dates(course_name, dates_file, active, dryrun):
    """Set due dates for assignments from a dates file.

    Input format: assignment name TAB comma-separated dates

    Each date is type=YYYY-MM-DD-hh:mm where type is available, due, or until.

    Example: Homework 1\tavailable=2024-01-15-09:00,due=2024-01-22-23:59
    """
    canvas = get_canvas_object()
    course = get_course(canvas, course_name, active)

    # Build assignment lookup by name
    assignments = {a.name: a for a in course.get_assignments()}

    for line in dates_file:
        line = line.strip()
        if not line:
            continue
        parts = line.split('\t')
        if len(parts) < 2:
            warn(f"skipping malformed line: {line}")
            continue

        name = parts[0]
        date_entries = parse_date_entries(parts[1])

        if name not in assignments:
            error(f"assignment not found: {name}")
            continue

        if not date_entries:
            info(f"no dates to set for: {name}")
            continue

        if dryrun:
            info(f"would update {name} with {date_entries}")
        else:
            info(f"updating {name}")
            assignments[name].edit(assignment=date_entries)



import datetime
from datetime import timezone

import click

from canvas_sak.core import *

namedtuple("ConversationMeta", ["id", ])

@canvas_sak.command()
@click.argument('course_substring')
@click.option('--dryrun/--no-dryrun', default=True, help="show what would be done, but don't do it")
def archive_inbox(course_substring, dryrun):
    '''
    move inbox conversations for a course to the archive.
    the course can be a partial name or * for all courses.
    '''
    recent_past = datetime.datetime.now() - datetime.timedelta(minutes=60)
    canvas_recent_past = recent_past.strftime('%Y-%m-%dT%H:%M:%SZ')
    canvas = get_canvas_object()
    to_archive = []
    archive_count = 0
    skipped_count = 0
    course_substring = course_substring.lower()
    archive_contexts = set()
    for c in canvas.get_conversations():
        if c.last_message_at > canvas_recent_past:
            continue

        if c.context_name and (course_substring == '*' or course_substring in c.context_name.lower()):
            archive_count += 1
            to_archive = to_archive + [c.id]
            archive_contexts.add(c.context_name)
        else:
            skipped_count += 1
        click.echo(f"\rto archive {archive_count} skipped {skipped_count}   ", nl=False)
    click.echo()
    sorted_contexts = list(archive_contexts)
    sorted_contexts.sort()
    if sorted_contexts:
        click.echo("archiving conversations for these contexts:")
    for context in sorted_contexts:
        click.echo(context)
    if dryrun:
        click.echo("this was a dry run, so no conversations were archived")
    else:
        if to_archive:
            click.echo(f"archiving {len(to_archive)} conversations")
            chunk_size = 100
            for r in range(0, len(to_archive), chunk_size):
                chunk = to_archive[r:r+chunk_size]
                click.echo (f"archiving {len(chunk)} conversations")
                canvas.conversations_batch_update(conversation_ids=chunk, event='archive')
        else:
            click.echo("no conversations to archive")


"""Tests for header parsing in upload_canvas_course.py"""

import pytest
from canvas_sak.commands.upload_canvas_course import (
    parse_headers,
    PAGE_KEYWORDS,
    DISCUSSION_KEYWORDS,
)


class TestParseHeaders:
    """Test cases for the parse_headers function."""

    def test_valid_keywords_parsed_correctly(self):
        """Valid keywords should be extracted into the headers dict."""
        content = "published: true\ntitle: My Page\nThis is the body content."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['published'] == 'true'
        assert headers['title'] == 'My Page'
        assert body == 'This is the body content.\n'

    def test_title_does_not_stop_header_parsing(self):
        """Parsing should continue after title keyword to capture other keywords."""
        content = "title: Test Title\npublished: true\nBody text here."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['title'] == 'Test Title'
        assert headers['published'] == 'true'  # Should be parsed, not in body
        assert body == 'Body text here.\n'

    def test_unknown_keyword_treated_as_content(self):
        """Unknown keywords should be treated as normal content, not discarded."""
        content = "unknown_key: some value\ntitle: My Title\nBody content."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        # Unknown keyword line should be preserved in body
        assert 'unknown_key' not in headers
        assert body == 'unknown_key: some value\ntitle: My Title\nBody content.'

    def test_unknown_keyword_preserves_entire_line(self):
        """The entire line with unknown keyword should be preserved in body."""
        content = "invalid: this should be in body\nMore body text."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers == {}
        assert body == 'invalid: this should be in body\nMore body text.'

    def test_all_page_keywords_recognized(self):
        """All PAGE_KEYWORDS should be recognized and parsed."""
        content = "published: true\npublish_at: 2024-01-01\nfront_page: true\ntitle: Full Page\nBody."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['published'] == 'true'
        assert headers['publish_at'] == '2024-01-01'
        assert headers['front_page'] == 'true'
        assert headers['title'] == 'Full Page'
        assert body == 'Body.\n'

    def test_all_discussion_keywords_recognized(self):
        """All DISCUSSION_KEYWORDS should be recognized and parsed."""
        content = "published: false\npublish_at: 2024-06-15\ntitle: Discussion Topic\nMessage body."
        headers, body = parse_headers(content, DISCUSSION_KEYWORDS)

        assert headers['published'] == 'false'
        assert headers['publish_at'] == '2024-06-15'
        assert headers['title'] == 'Discussion Topic'
        assert body == 'Message body.\n'

    def test_discussion_does_not_recognize_page_only_keywords(self):
        """DISCUSSION_KEYWORDS should not include page-only keywords like front_page."""
        content = "front_page: true\ntitle: My Discussion\nBody."
        headers, body = parse_headers(content, DISCUSSION_KEYWORDS)

        # front_page is not in DISCUSSION_KEYWORDS, so it should be treated as content
        assert 'front_page' not in headers
        assert body == 'front_page: true\ntitle: My Discussion\nBody.'

    def test_colon_in_value_preserved(self):
        """Colons within the value portion should be preserved."""
        content = "title: Time: 10:30 AM\nBody content."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['title'] == 'Time: 10:30 AM'
        assert body == 'Body content.\n'

    def test_whitespace_stripped_from_values(self):
        """Whitespace should be stripped from header values."""
        content = "title:   Lots of spaces   \nBody."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['title'] == 'Lots of spaces'

    def test_empty_body_after_keywords(self):
        """Should handle case where there's no body content after keywords."""
        content = "title: Just a Title"
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['title'] == 'Just a Title'
        assert body == '\n'  # Empty line after headers becomes body

    def test_multiline_body_preserved(self):
        """Multiline body content should be fully preserved."""
        content = "title: My Page\nLine 1\nLine 2\nLine 3"
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['title'] == 'My Page'
        assert body == 'Line 1\nLine 2\nLine 3'

    def test_line_without_colon_treated_as_content(self):
        """Lines without colons should be treated as content."""
        content = "This line has no colon\ntitle: My Title\nBody."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        # Line without colon should start the body
        assert 'title' not in headers
        assert body == 'This line has no colon\ntitle: My Title\nBody.'

    def test_empty_content(self):
        """Should handle empty content gracefully."""
        content = ""
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers == {}
        assert body == '\n'  # partition on empty string gives ('', '', '')

    def test_only_title_keyword(self):
        """Should work with only the title keyword and body."""
        content = "title: Simple\nJust body text here."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers == {'title': 'Simple'}
        assert body == 'Just body text here.\n'

    def test_case_sensitive_keywords(self):
        """Keywords should be case-sensitive."""
        content = "Title: Uppercase T\ntitle: Lowercase t\nBody."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        # 'Title' (capital T) is not in PAGE_KEYWORDS, so treated as content
        assert 'Title' not in headers
        assert body == 'Title: Uppercase T\ntitle: Lowercase t\nBody.'

    def test_custom_keywords_set(self):
        """Should work with custom keyword sets."""
        custom_keywords = {'foo', 'bar', 'title'}
        content = "foo: value1\nbar: value2\ntitle: Custom\nBody."
        headers, body = parse_headers(content, custom_keywords)

        assert headers['foo'] == 'value1'
        assert headers['bar'] == 'value2'
        assert headers['title'] == 'Custom'
        assert body == 'Body.\n'

    def test_unknown_keyword_after_valid_keywords(self):
        """Unknown keyword after valid ones should stop parsing and preserve remaining content."""
        content = "published: true\nunknown: value\ntitle: Never Reached\nBody."
        headers, body = parse_headers(content, PAGE_KEYWORDS)

        assert headers['published'] == 'true'
        assert 'unknown' not in headers
        assert 'title' not in headers
        assert body == 'unknown: value\ntitle: Never Reached\nBody.'



from canvas_sak.core import *

@canvas_sak.command()
@click.argument("course")
@click.option('-t', 'thresholds', metavar='threshold', multiple=True, default=[84, 90, 95], show_default=True,
              type=click.INT, help="""
              assignment groups with grades about the lowest threshold will have a +, the next
              lowest gets ++, and so on. assignment groups below the lowest threshold will not
              be printed.
              """)
@click.option('-s', 'skip', metavar='skip_assignment', multiple=True, default=['iclickr', 'ungraded', 'imported'],
              show_default=True, help="""
              assignment groups with the listed keywords will not be collected.
              """)
def collect_reference_info(course, thresholds, skip):
    '''collect high level information about students of previous classes to help writing reference letters'''
    Grade = namedtuple('Grade', ['category', 'grade'])
    canvas = get_canvas_object()
    for course in get_courses(canvas, course, is_active=False, is_finished=True):
        results = canvas.graphql('query { course(id: "' + str(course.id) + '''") {
              assignmentGroupsConnection {
                  nodes { name
                          gradesConnection {
                              edges { node { currentScore
                                             enrollment { user { name } }
                                           } } } } } } }''')
        grades_by_student = defaultdict(list)

        for assignment_group in results['data']['course']['assignmentGroupsConnection']['nodes']:
            category = assignment_group['name']
            if any(x in category.lower() for x in skip):
                continue
            for grade in assignment_group['gradesConnection']['edges']:
                score = grade['node']['currentScore']
                name = grade['node']['enrollment']['user']['name']
                if score:
                    pluses = to_plus(score, thresholds)
                    if pluses:
                        grades_by_student[name].append(Grade(category, pluses))
        for i in grades_by_student.items():
            label = f'{i[0]}@{format_course_name(course.name)}'
            output(f'{label} {" ".join([g.category + ":" + g.grade for g in i[1]])}')

def to_plus(grade, levels):
    """ convert a score to a list of pluses based on grade """
    pluses = ""
    for level in levels:
        if grade >= level:
            pluses += '+'
    return pluses





from canvas_sak.core import *

@canvas_sak.command()
@click.argument('course')
@click.option('--active/--inactive', default=True, help="show only active courses")
@click.option('--emails/--no-emails', help="list student emails")
@click.option('--id/--no-id', help="include the canvas id")
@click.option('--link', help="show value of a link field (* for everything)", default=None)
def list_students(course, active, emails, link, id):
    '''list the students in a course'''
    if link:
        link = link.lower()
    canvas = get_canvas_object()
    course = get_course(canvas, course, active)
    users = course.get_users(include=["enrollments"])
    for user in users:
        initial_info = ""
        additional_info = ""
        if id:
            initial_info += f"{user.login_id}\t"
        if emails or link:
            profile = user.get_profile(include=["links"])
            additional_info += f"\t{profile['primary_email'] if emails else ''}"
            if link:
                link_info = "\t" + " ".join([f"{m['title']}={m['url']}" for m in profile['links'] if m['title'].lower() == link or link == '*'])
                additional_info += link_info
        output(f"{initial_info}{user.name}{additional_info}")




import click

from canvas_sak.core import *

# from https://stackoverflow.com/a/925630
from io import StringIO
from html.parser import HTMLParser

class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs= True
        self.text = StringIO()
    def handle_data(self, d):
        self.text.write(d)
    def get_data(self):
        return self.text.getvalue()

def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()

def dehtml(s):
    if type(s) is not str:
        return s
    return strip_tags(s.replace("&nbsp;", "").strip().replace("\n", " ").replace("\r",""))


def evolves(base, change):
    if type(base) is not str or type(change) is not str:
        return True
    base_words = set(base.split())
    change_words = set(change.split())
    added = change_words - base_words
    removed = base_words - change_words
    rc = True
    if len(added) == 0:
        rc = False
    if len(added) == 1 and len(removed) >= 1:
        a = added.pop()
        for r in removed:
            if r.startswith(a):
                rc = False
                break
    return rc


def get_question_group(quiz, question_groups, group_id):
    if group_id not in question_groups:
        question_groups[group_id] = quiz.get_quiz_group(group_id)
    return question_groups[group_id]


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('quiz_name', metavar='quiz', default='')
@click.option('--show-question/--no-show-question', default=False, show_default=True)
@click.option('--for-student', metavar='students', default=[], multiple=True,
              help='students to get quiz logs for')
@click.option('--summarize/--no-summarize', default=True, show_default=True,
              help='show only completed answers. skip answers that are a prefix of subsequent answers')
@click.option('--final-answer/--no-final-answer', default=True, show_default=True,
              help='show the final answer, if --no-final-answer, the final answers will be skipped')
def quiz(course_name, quiz_name, show_question, for_student, summarize, final_answer):
    '''
    get quiz logs for a student
    '''
    canvas = get_canvas_object()

    course = get_course(canvas, course_name)

    students = [s.lower() for s in for_student]

    users = {u.id : u.name for u in course.get_users() if len(students) == 0 or [s for s in students if s in u.name.lower()]}
    if len(users) == 0:
        error(f"no students matched {students}")
        exit(2)

    quiz = [q for q in course.get_quizzes() if quiz_name in q.title]
    if len(quiz) == 0:
        error(f"no quizzes matched {quiz_name}")
        exit(2)
    if len(quiz) > 1:
        error(f"multiple matches for {quiz_name}")
        for q in quiz:
            error(f"    {q.title}")
        exit(2)

    quiz = quiz[0]

    answers = []

    questions = {q.id: q for q in quiz.get_questions()}
    question_groups = {}
    for s in quiz.get_submissions():
        if s.user_id not in users:
            continue
        prev = None
        for es in s.get_submission_events():
            time_spent = int(s.time_spent)
            if es.event_type != 'question_answered':
                continue
            event_data = es.event_data
            if not type(event_data) is list:
                event_data = [event_data]
            for e in event_data:
                if e['answer']:
                    question = questions[int(e['quiz_question_id'])]
                    if question.quiz_group_id:
                        position = f"{get_question_group(quiz, question_groups, question.quiz_group_id).position:2}.{question.position:2}"
                    else:
                        position = f"{question.position:5}"
                    data = (users[s.user_id], position, f"{time_spent//60:02}:{time_spent%60:02}", dehtml(e['answer']), question.question_text)
                    answers.append(data)

    if summarize or final_answer:
        answers.sort(reverse=True)
        prev=None
        new_answers = []
        final_answers = {}
        for a in answers:
            skip_answer = False
            if summarize and prev != None and prev[0] == a[0] and prev[1] == a[1] and not evolves(prev[3], a[3]):
                skip_answer = True
            if final_answer and not skip_answer:
                final_key = (a[0], a[1])
                if final_key not in final_answers:
                    final_answers[final_key] = a
                    skip_answer = True
            if not skip_answer:
                new_answers.append(a)
            prev = a
        answers = new_answers

    answers.sort(key=lambda x: x[2])
    for a in answers:
        print(f"{a[1]} {a[2]} {a[0]} {a[3]} {a[4] if show_question else ''}")



from canvas_sak.core import *

@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.argument('assignment_name', metavar='assignment', default='')
@click.option('--dryrun/--no-dryrun', default=True, show_default=True,
              help="only show the grade, don't actually set it")
def download_submissions(course_name, assignment_name, dryrun):
    '''
    download submissions for an assignment.
    '''

    canvas = get_canvas_object()

    course = get_course(canvas, course_name)

    assignment = get_assignment(course, assignment_name)

    query = """
    query submissions($assignmentid: ID!) {
        assignment(id: $assignmentid) { submissionsConnection {
            nodes { attachments { url displayName } user { name }
                    commentsConnection { nodes { comment attachments { url displayName}}}
            }
        }}
    }
    """
    result = canvas.graphql(query, {"assignmentid": assignment.id})
    submissions = result['data']['assignment']['submissionsConnection']['nodes']

    if dryrun:
        info(f"{len(submissions)} submissions to download")
        sys.exit(0)

    with click.progressbar(length=len(submissions), label="downloading submission", show_pos=True) as bar:
        for s in submissions:
            count = 1
            name = s['user']['name']
            dir = os.path.join(assignment_name, name.replace(' ', '-'))
            os.makedirs(dir, exist_ok=True)
            for a in s['attachments']:
                download_attachment(f'{dir}/submission{count}', a)
                count += 1
            count = 1
            for c in s['commentsConnection']['nodes']:
                with open(os.path.join(dir, f"comment{count}.txt"), 'w') as fd:
                    fd.write(c['comment'])
                subcount = 1
                for ca in c['attachments']:
                    download_attachment(f'{dir}/comment{count}attachment{subcount}', ca)
                    subcount += 1
                count += 1
            bar.update(1)



def download_attachment(basename, a):
    fname = a['displayName']
    suffix = os.path.splitext(fname)[1]
    durl = a['url']
    info(f'downloading {a}')
    with requests.get(durl) as response:
        if response.status_code != 200:
            error(f'error {response.status_code} fetching {durl}')
            return
        with open(f"{basename}{suffix}", "wb") as fd:
            for chunk in response.iter_content():
                fd.write(chunk)



import markdownify
import markdown


def html2mdstr(html_str: str):
    """Converts html in string form to markdown"""
    md_str = markdownify.markdownify(html_str)
    return md_str


def html2mdlist(html_list: list):
    """Converts html as a list of strings to markdown"""
    html_str = '\n'.join(html_list)
    return html2mdstr(html_str)


def md2htmlstr(md_str: str):
    """Converts markdown in string form to html"""
    html_str = markdown.markdown(md_str)
    return html_str


def md2htmllist(md_list: list):
    """Converts markdown as a list of strings to html"""
    md_str = '\n'.join(md_list)
    return md2htmlstr(md_str)



from canvasapi.page import Page

from canvas_sak.core import *
from canvas_sak.core import filter_ignored_paths
from canvas_sak.md2fhtml import md2htmlstr


def boolean_option(key, params):
    return params[key].lower() == "true" if key in params else False


def create_assignment(course, name, description=""):
    rc = course.create_assignment({"name": name, "description": description})
    process_resource_record(ResourceRecord(rc.id, base_url(rc.html_url), "Assignment", name, not description))
    return rc


def create_discussion(course, name, message=""):
    rc = course.create_discussion_topic(title=name, message=message)
    process_resource_record(ResourceRecord(rc.id, base_url(rc.html_url), "Discussion", name, not message))
    return rc


class FileStub:
    """Simple wrapper to provide .id attribute for uploaded files."""
    def __init__(self, file_id, url):
        self.id = file_id
        self.url = url


def create_file(course, name, content=b' '):
    last_slash = name.rindex("/")
    file = name[last_slash + 1:]
    parent = name[0:last_slash] if last_slash != -1 else ""
    # course.upload expects a file path, not bytes - use a temp file
    with tempfile.NamedTemporaryFile(delete=False, suffix=file) as tmp:
        tmp.write(content)
        tmp_path = tmp.name
    try:
        success, response = course.upload(tmp_path, parent_folder_path=parent, name=file)
    finally:
        os.unlink(tmp_path)
    # response can be a dict or a File object depending on canvasapi version
    if isinstance(response, dict):
        file_id = response['id']
        file_url = response['url']
    else:
        file_id = response.id
        file_url = response.url
    process_resource_record(ResourceRecord(file_id, base_url(file_url), "File", name, not content))
    return FileStub(file_id, file_url)


def create_quiz(course, name, description=''):
    rc = course.create_quiz({"title": name, "description": description})
    process_resource_record(ResourceRecord(rc.id, base_url(rc.html_url), "Quiz", name, not description))
    return rc


def create_stub(course, item_type, item_name):
    if item_type == "Assignment":
        return create_assignment(course, item_name)
    if item_type == "Discussion":
        return create_discussion(course, item_name)
    if item_type == "File":
        return create_file(course, item_name)
    if item_type == "Quiz":
        return create_quiz(course, item_name)


def create_page(course, title, body=None):
    rrname = "Page" + title
    if rrname in rr4name:
        return rr4name[rrname]
    page_dict = {"title": title}
    if body:
        page_dict["body"] = body
    rc = course.create_page(page_dict)
    process_resource_record(ResourceRecord(rc.page_id, rc.url, "Page", title, not body))
    return rc


def upload_modules(course, source, dryrun):
    last_module_seen = None
    last_module_item_names = set()
    with open(source, "r") as fd:
        for line in fd.readlines():
            m = re.match(r"^((  )+) ?\*\s+([^;]+)(;(.*)$)?", line)
            if line.startswith("# "):
                parts = line[2:].strip().split(";", 2)
                title = parts[0]
                if title in course_modules:
                    last_module_seen = course_modules[title]
                    last_module_item_names = set(
                        [f"{mi.type}; {mi.title}" for mi in course_modules[title].get_module_items()])
                    info(f"{title} module already present")
                elif dryrun:
                    info(f"would create {title} module")
                else:
                    info(f"creating {title} module")
                    last_module_seen = course.create_module({"name": title, "published": boolean_option("published",
                                                                                                        extract_options(
                                                                                                            parts[
                                                                                                                1])) if len(
                        parts) > 1 else True})
                    course_modules[title] = last_module_seen
                    last_module_item_names = set()
            elif m:
                indent_level = len(m.group(1)) / 2
                item_title = m.group(3)
                item_parts = m.group(5).split(';', 1)
                item_options = extract_options(item_parts[1]) if len(item_parts) > 1 else {}
                item_type = item_parts[0].strip()
                item_key = f"{item_type}; {item_title}"
                if item_key in last_module_item_names:
                    info(f"item {item_key} present in {title}")
                elif dryrun:
                    info(f"would create item {item_title} in {title}")
                else:
                    info(f"creating {item_title} in {title}")
                    item_dict = {"title": item_title, "indent": indent_level - 1, "type": item_type}
                    if "newtab" in item_options:
                        item_dict["new_tab"] = boolean_option("newtab", item_options)
                    item_name = item_options["target"] if "target" in item_options else item_title
                    if item_type in ["Assignment", "Discussion", "File", "Quiz"]:
                        item_name = item_options["target"] if "target" in item_options else item_title
                        name_key = item_type + item_name
                        item_dict["content_id"] = rr4name[name_key].id if name_key in rr4name else create_stub(course,
                                                                                                               item_type,
                                                                                                               item_name).id
                    elif item_type == "Page":
                        item_page = create_page(course, item_name)
                        url = item_page.url
                        item_dict["page_url"] = url
                    elif item_type.startswith("External"):
                        item_dict["external_url"] = item_options["url"]
                    if item_type == "ExternalTool":
                        error("ExternalTool creation not currently supported")
                    else:
                        last_module_seen.create_module_item(item_dict)
                        last_module_item_names.add(item_key)


def page_name_to_url(item_name):
    return item_name.lower().replace(" ", "-").replace(":", "").replace("--", '-')


def extract_options(options):
    return {s[0].strip().lower(): s[1].strip() if len(s) > 1 else "" for s in
            [o.split('=', 2) for o in options.split(';')]}


DISCUSSION_KEYWORDS = set(["title", "published", "publish_at"])


def parse_headers(content, keywords):
    """Parse header lines from content until an unknown keyword is encountered.

    Returns a tuple of (headers_dict, remaining_body).
    Continues processing all recognized keywords until a line doesn't match,
    then treats that line and everything after as normal content.
    """
    headers = {}
    page = content
    while True:
        (line, _, page) = page.partition('\n')
        (key, _, value) = line.partition(':')
        if key in keywords:
            headers[key] = value.strip()
        else:
            # Unknown keyword - treat line as normal content
            page = line + '\n' + page
            break
    return headers, page


def upload_discussions(course, source, dryrun, force):
    all_files = [os.path.join(d, f)[len(source) + 1:].replace("\\", "/") for (d, sds, fs) in os.walk(source) for f in fs]
    to_upload = set(filter_ignored_paths(all_files))
    for file in to_upload:
        with open(os.path.join(source, file), "r") as fd:
            page = fd.read()
        dict, page = parse_headers(page, DISCUSSION_KEYWORDS)
        dict['message'] = md2htmlstr(page)
        dict['discussion_type'] = 'threaded'
        rrkey = "Discussion" + dict['title']
        exists = rrkey in rr4name
        if exists:
            info(f"discussion {dict['title']} already exists")
        do_upload = not exists if not force else True
        if do_upload:
            if exists:
                if dryrun:
                    info(f"would update {dict['title']} from {file}")
                else:
                    info(f"updating {dict['title']} from {file}")
                    dict['title'] = dict['title'] + "blah"
                    dt = course.get_discussion_topic(rr4name[rrkey].id)
                    rc = dt.update(**dict)
            else:
                if dryrun:
                    info(f"would create {dict['title']} from {file}")
                else:
                    info(f"creating {dict['title']} from {file}")
                    rc = course.create_discussion_topic(**dict)
                    process_resource_record(ResourceRecord(rc.id, base_url(rc.html_url), "Discussion", rc.title, False))


def upload_assignments(course, target, dryrun):
    pass


PAGE_KEYWORDS = set(["title", "published", "publish_at", "front_page"])


def upload_pages(course, source, dryrun, force):
    # got to watch out for windows \\ when using join!
    all_files = [os.path.join(d, f)[len(source) + 1:].replace("\\", "/") for (d, sds, fs) in os.walk(source) for f in fs]
    to_upload = set(filter_ignored_paths(all_files))
    for file in to_upload:
        with open(os.path.join(source, file), "r") as fd:
            page = fd.read()
        dict, page = parse_headers(page, PAGE_KEYWORDS)
        dict['body'] = md2htmlstr(page)
        rrkey = "Page" + dict['title']
        exists = rrkey in rr4name
        if exists:
            info(f"page {dict['title']} already exists")
        do_upload = not exists if not force else True
        if do_upload:
            if exists:
                if dryrun:
                    info(f"would update {dict['title']} from {file}")
                else:
                    info(f"updating {dict['title']} from {file}")
                    dict['title'] = dict['title'] + "blah"
                    p: Page = course.get_page(rr4name[rrkey].url)
                    rc = p.edit(**dict)
            else:
                if dryrun:
                    info(f"would create {dict['title']} from {file}")
                else:
                    info(f"creating {dict['title']} from {file}")
                    rc = course.create_page(dict)
                    process_resource_record(ResourceRecord(rc.page_id, rc.url, "Page", rc.title, False))


def upload_files(course, target, dryrun):
    # got to watch out for windows \\ when using join!
    all_files = [os.path.join(d, f)[len(target) + 1:].replace("\\", "/") for (d, sds, fs) in os.walk(target) for f in fs]
    to_upload = set(filter_ignored_paths(all_files))

    existing_files = set()
    for folder in course.get_folders():
        for file in folder.get_files():
            existing_files.add(os.path.join(str(folder), str(file)).replace("\\", "/"))
    for common in to_upload.intersection(existing_files):
        warn(f"{common} already exists. skipping.")

    uploads = list(to_upload.difference(existing_files))
    if dryrun:
        for up in uploads:
            name = os.path.basename(up)
            parent = os.path.dirname(up)
            info(f"would upload {name} to {parent}")
    else:
        with click.progressbar(uploads, label="uploading",
                               item_show_func=lambda i: i if i else "") as ups:
            for up in ups:
                name = os.path.basename(up)
                parent = os.path.dirname(up)
                filename = os.path.join(target, up)
                if os.stat(filename).st_size > 0:
                    course.upload(os.path.join(target, up), parent_folder_path=parent, name=name)


def upload_announcements(course, target, dryrun):
    pass


@canvas_sak.command()
@click.argument('course_name', metavar='course')
@click.option('--dryrun/--no-dryrun', default=True, show_default=True, help="show what would happen, but don't do it.")
@click.option('--modules/--no-modules', default=False, show_default=True,
              help=f"upload modules to the {click.style('modules', underline=True, italic=True)} subdirectory.")
@click.option('--discussions/--no-discussions', default=False, show_default=True,
              help=f"upload discussions to the {click.style('discussions', underline=True, italic=True)} subdirectory.")
@click.option('--assignments', default=False, show_default=True,
              help=f"upload assignments to the {click.style('assignments', underline=True, italic=True)} subdirectory.")
@click.option('--pages/--no-pages', default=False, show_default=True,
              help=f"upload pages to the {click.style('pages', underline=True, italic=True)} subdirectory.")
@click.option('--files/--no-files', default=False, show_default=True,
              help=f"upload files to the {click.style('files', underline=True, italic=True)} subdirectory.")
@click.option('--announcements', default=False, show_default=True,
              help=f"upload announcements to the {click.style('announcements', underline=True, italic=True)} subdirectory.")
@click.option('--all/--no-all', default=False, show_default=True,
              help="upload all content to corresponding directories")
@click.option("--source", default='.', show_default=True, help="upload content parent directory.")
@click.option("--force/--no-force", default=False, show_default=True, help="overwrite existing content")
def upload_course_content(course_name, dryrun, modules, discussions, assignments, pages, files, announcements, all,
                          source, force):
    """upload course content from local files"""
    canvas = get_canvas_object()
    course = get_course(canvas, course_name, is_active=False)
    output(f"found {course.name}")
    map_course_resource_records(course)

    if all:
        modules = discussions = assignments = pages = files = announcements = True

    if not (modules or discussions or assignments or pages or files or announcements):
        error("nothing selected to upload")
        exit(1)

    if discussions:
        upload_discussions(course, os.path.join(source, 'discussions'), dryrun, force)
    if assignments:
        upload_assignments(course, os.path.join(source, 'assignments'), dryrun)
    if pages:
        upload_pages(course, os.path.join(source, 'pages'), dryrun, force)
    if files:
        upload_files(course, os.path.join(source, 'files'), dryrun)
    if announcements:
        upload_announcements(course, os.path.join(source, 'announcements'), dryrun)
    if modules:
        upload_modules(course, os.path.join(source, 'modules'), dryrun)
